/* -*-c-*-
 *
 * The MD5 message digest function
 *
 * (c) 1998 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#include "ghash.h"
#include "ghash-def.h"
#include "hash.h"
#include "md5.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @md5_compress@ --- *
 *
 * Arguments:	@md5_ctx *ctx@ = pointer to context block
 *		@const void *sbuf@ = pointer to buffer of appropriate size
 *
 * Returns:	---
 *
 * Use:		MD5 compression function.
 */

void md5_compress(md5_ctx *ctx, const void *sbuf)
{
  uint32 a, b, c, d;
  uint32 buf[16];

  /* --- Fetch the chaining variables --- */

  a = ctx->a;
  b = ctx->b;
  c = ctx->c;
  d = ctx->d;

  /* --- Fetch the buffer contents --- */

  {
    int i;
    const octet *p;

    for (i = 0, p = sbuf; i < 16; i++, p += 4)
      buf[i] = LOAD32_L(p);
  }

  /* --- Definitions for round functions --- */

#define F(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & ~(z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | ~(z)))

#define T(w, x, y, z, i, r, k, f) do {					\
  uint32 _t = w + f(x, y, z) + buf[i] + k;				\
  w = ROL32(_t, r) + x;							\
} while (0)

#define FF(w, x, y, z, i, r, k) T(w, x, y, z, i, r, k, F)
#define GG(w, x, y, z, i, r, k) T(w, x, y, z, i, r, k, G)
#define HH(w, x, y, z, i, r, k) T(w, x, y, z, i, r, k, H)
#define II(w, x, y, z, i, r, k) T(w, x, y, z, i, r, k, I)

  /* --- The main compression function --- */

  FF(a, b, c, d,  0,  7, 0xd76aa478);
  FF(d, a, b, c,  1, 12, 0xe8c7b756);
  FF(c, d, a, b,  2, 17, 0x242070db);
  FF(b, c, d, a,  3, 22, 0xc1bdceee);
  FF(a, b, c, d,  4,  7, 0xf57c0faf);
  FF(d, a, b, c,  5, 12, 0x4787c62a);
  FF(c, d, a, b,  6, 17, 0xa8304613);
  FF(b, c, d, a,  7, 22, 0xfd469501);
  FF(a, b, c, d,  8,  7, 0x698098d8);
  FF(d, a, b, c,  9, 12, 0x8b44f7af);
  FF(c, d, a, b, 10, 17, 0xffff5bb1);
  FF(b, c, d, a, 11, 22, 0x895cd7be);
  FF(a, b, c, d, 12,  7, 0x6b901122);
  FF(d, a, b, c, 13, 12, 0xfd987193);
  FF(c, d, a, b, 14, 17, 0xa679438e);
  FF(b, c, d, a, 15, 22, 0x49b40821);

  GG(a, b, c, d,  1,  5, 0xf61e2562);
  GG(d, a, b, c,  6,  9, 0xc040b340);
  GG(c, d, a, b, 11, 14, 0x265e5a51);
  GG(b, c, d, a,  0, 20, 0xe9b6c7aa);
  GG(a, b, c, d,  5,  5, 0xd62f105d);
  GG(d, a, b, c, 10,  9, 0x02441453);
  GG(c, d, a, b, 15, 14, 0xd8a1e681);
  GG(b, c, d, a,  4, 20, 0xe7d3fbc8);
  GG(a, b, c, d,  9,  5, 0x21e1cde6);
  GG(d, a, b, c, 14,  9, 0xc33707d6);
  GG(c, d, a, b,  3, 14, 0xf4d50d87);
  GG(b, c, d, a,  8, 20, 0x455a14ed);
  GG(a, b, c, d, 13,  5, 0xa9e3e905);
  GG(d, a, b, c,  2,  9, 0xfcefa3f8);
  GG(c, d, a, b,  7, 14, 0x676f02d9);
  GG(b, c, d, a, 12, 20, 0x8d2a4c8a);

  HH(a, b, c, d,  5,  4, 0xfffa3942);
  HH(d, a, b, c,  8, 11, 0x8771f681);
  HH(c, d, a, b, 11, 16, 0x6d9d6122);
  HH(b, c, d, a, 14, 23, 0xfde5380c);
  HH(a, b, c, d,  1,  4, 0xa4beea44);
  HH(d, a, b, c,  4, 11, 0x4bdecfa9);
  HH(c, d, a, b,  7, 16, 0xf6bb4b60);
  HH(b, c, d, a, 10, 23, 0xbebfbc70);
  HH(a, b, c, d, 13,  4, 0x289b7ec6);
  HH(d, a, b, c,  0, 11, 0xeaa127fa);
  HH(c, d, a, b,  3, 16, 0xd4ef3085);
  HH(b, c, d, a,  6, 23, 0x04881d05);
  HH(a, b, c, d,  9,  4, 0xd9d4d039);
  HH(d, a, b, c, 12, 11, 0xe6db99e5);
  HH(c, d, a, b, 15, 16, 0x1fa27cf8);
  HH(b, c, d, a,  2, 23, 0xc4ac5665);

  II(a, b, c, d,  0,  6, 0xf4292244);
  II(d, a, b, c,  7, 10, 0x432aff97);
  II(c, d, a, b, 14, 15, 0xab9423a7);
  II(b, c, d, a,  5, 21, 0xfc93a039);
  II(a, b, c, d, 12,  6, 0x655b59c3);
  II(d, a, b, c,  3, 10, 0x8f0ccc92);
  II(c, d, a, b, 10, 15, 0xffeff47d);
  II(b, c, d, a,  1, 21, 0x85845dd1);
  II(a, b, c, d,  8,  6, 0x6fa87e4f);
  II(d, a, b, c, 15, 10, 0xfe2ce6e0);
  II(c, d, a, b,  6, 15, 0xa3014314);
  II(b, c, d, a, 13, 21, 0x4e0811a1);
  II(a, b, c, d,  4,  6, 0xf7537e82);
  II(d, a, b, c, 11, 10, 0xbd3af235);
  II(c, d, a, b,  2, 15, 0x2ad7d2bb);
  II(b, c, d, a,  9, 21, 0xeb86d391);

  /* --- Update the chaining variables --- */

  ctx->a += a;
  ctx->b += b;
  ctx->c += c;
  ctx->d += d;
}

/* --- @md5_init@ --- *
 *
 * Arguments:	@md5_ctx *ctx@ = pointer to context block to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes a context block ready for hashing.
 */

void md5_init(md5_ctx *ctx)
{
  ctx->a = 0x67452301;
  ctx->b = 0xefcdab89;
  ctx->c = 0x98badcfe;
  ctx->d = 0x10325476;
  ctx->off = 0;
  ctx->nl = ctx->nh = 0;
}

/* --- @md5_set@ --- *
 *
 * Arguments:	@md5_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = pointer to state buffer
 *		@unsigned long count@ = current count of bytes processed
 *
 * Returns:	---
 *
 * Use:		Initializes a context block from a given state.  This is
 *		useful in cases where the initial hash state is meant to be
 *		secret, e.g., for NMAC and HMAC support.
 */

void md5_set(md5_ctx *ctx, const void *buf, unsigned long count)
{
  const octet *p = buf;
  ctx->a = LOAD32_L(p +	 0);
  ctx->b = LOAD32_L(p +	 4);
  ctx->c = LOAD32_L(p +	 8);
  ctx->d = LOAD32_L(p + 12);
  ctx->off = 0;
  ctx->nl = U32(count);
  ctx->nh = U32(((count & ~(unsigned long)MASK32) >> 16) >> 16);
}

/* --- @md5_hash@ --- *
 *
 * Arguments:	@md5_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = buffer of data to hash
 *		@size_t sz@ = size of buffer to hash
 *
 * Returns:	---
 *
 * Use:		Hashes a buffer of data.  The buffer may be of any size and
 *		alignment.
 */

void md5_hash(md5_ctx *ctx, const void *buf, size_t sz)
{
  HASH_BUFFER(MD5, md5, ctx, buf, sz);
}

/* --- @md5_done@ --- *
 *
 * Arguments:	@md5_ctx *ctx@ = pointer to context block
 *		@void *hash@ = pointer to output buffer
 *
 * Returns:	---
 *
 * Use:		Returns the hash of the data read so far.
 */

void md5_done(md5_ctx *ctx, void *hash)
{
  octet *p = hash;
  HASH_MD5STRENGTH(MD5, md5, ctx);
  STORE32_L(p +	 0, ctx->a);
  STORE32_L(p +	 4, ctx->b);
  STORE32_L(p +	 8, ctx->c);
  STORE32_L(p + 12, ctx->d);
}

/* --- @md5_state@ --- *
 *
 * Arguments:	@md5_ctx *ctx@ = pointer to context
 *		@void *state@ = pointer to buffer for current state
 *
 * Returns:	Number of bytes written to the hash function so far.
 *
 * Use:		Returns the current state of the hash function such that
 *		it can be passed to @md5_set@.
 */

unsigned long md5_state(md5_ctx *ctx, void *state)
{
  octet *p = state;
  STORE32_L(p +	 0, ctx->a);
  STORE32_L(p +	 4, ctx->b);
  STORE32_L(p +	 8, ctx->c);
  STORE32_L(p + 12, ctx->d);
  return (ctx->nl | ((ctx->nh << 16) << 16));
}

/* --- Generic interface --- */

GHASH_DEF(MD5, md5)

/* --- Test code --- */

HASH_TEST(MD5, md5)

/*----- That's all, folks -------------------------------------------------*/
